#!/usr/bin/env python
#--------------------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2008
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program License Agreement.
# See the software licensing materials that came with the IBM XIV Storage System for terms and conditions.
#
#--------------------------------------------------------------------------------------------------------------

import host_attach
from host_attach import logger, is_error_free, execute_and_log, PACKAGE_DIR
from platform import platform
from sys import exit

GETUID_CALLOUT_PATH = PACKAGE_DIR + '/bin/xiv_uid_callout'
MULTIPATH_XIV_PATH = PACKAGE_DIR + '/src/multipath.conf.xiv'

def multipath_conf_backup():
	output = execute_and_log('''cp /etc/multipath.conf{,.`date +%Y%m%d-%H:%M:%S`}''')
	return 1

def multipath_conf_install_now():
	# needs to be more sophisticated
	output = execute_and_log('cp ' + MULTIPATH_XIV_PATH + ' /etc/multipath.conf')
	if not is_error_free(output):
		return 0
	return 1

def multipath_conf_install_later():
	output = execute_and_log('cp ' + MULTIPATH_XIV_PATH + ' /etc/multipath.conf.xiv')
	if not is_error_free(output):
		return 0
	return 1
	
def is_multipathconf_exist():
	output = execute_and_log('cat /etc/multipath.conf | head -n 1')
	if not is_error_free(output):
		return 0
	if output == '':
		return 0
	return 1
		
def multipath_conf_install():
	if not multipath_conf_backup():
		return 0
	if is_multipathconf_exist():
		etc_multipath = execute_and_log('cat /etc/multipath.conf')
		if not is_error_free(etc_multipath):
			logger.debug('failed to read /etc/multipath.conf')
	        original_multipath = execute_and_log('cat ' + PACKAGE_DIR + '/multipath.conf.original')
		if not is_error_free(original_multipath):
			logger.debug('failed to load ' + PACKAGE_DIR + '/multipath.conf.original')
		if original_multipath == etc_multipath:
			return multipath_conf_install_now()
		logger.debug('/etc/multipath.conf exists')
		if multipath_conf_install_later():
			logger.warning('/etc/multipath.conf already exists, skipping this configuration file')
			logger.warning('the needed configuration is in /etc/multipath.conf.xiv, merge it into /etc/multipath.conf')
			logger.info('run this script with --verify to make sure multipath is properly configured')
		else:
			logger.error('failed to configure multipath')
			return 0
	else:
		if not multipath_conf_install_now():
			return 0		
	return 1
	
def udev_install_rule(number=99, name='', rulesList=[]):
	output = execute_and_log('rm -rf /etc/udev/rules.d/' + number.__str__() + '-' + name + '.rules')
	if not is_error_free(output):
		return 0
	output = execute_and_log('touch /etc/udev/rules.d/' + number.__str__() + '-' + name + '.rules')
	if not is_error_free(output):
		return 0
	ruleFile = open('/etc/udev/rules.d/' + number.__str__() + '-' + name + '.rules', 'w')
	for rule in rulesList:
		ruleFile.write(rule + "\n")
		ruleFile.flush()
	ruleFile.close()
	return 1
	
def udev_install_rule_45():
	rules = []
	rules.append('''SUBSYSTEM=="block", ACTION=="add", KERNEL=="sd*[!0-9]", SYSFS{device/vendor}=="XIV", RUN+="/bin/sh -c 'echo 127 >  /sys/\$devpath/device/queue_depth'"''')
	rules.append('''SUBSYSTEM=="fc_remote_ports", ACTION=="add", SYSFS{port_name}=="0x1000001738*", RUN+="/bin/sh -c 'echo 1 >  /sys/\$devpath/dev_loss_tmo'"''')
	if not udev_install_rule(number=45, name='ibm-xiv', rulesList=rules):
		return 0
	return 1
	
def udev_install_rule_40():
	rules = []
	rules.append("# multipath wants the devmaps presented as meaninglful device names")
	rules.append("# so name them after their devmap name")
	rules.append('''SUBSYSTEM!="block", GOTO="end_mpath''')
	rules.append('''KERNEL!="dm-[0-9]*", ACTION=="add", PROGRAM=="/bin/bash -c '/sbin/lsmod | /bin/grep ^dm_multipath'", RUN+="/sbin/multipath -v0 %M:%m"''')
	rules.append('''KERNEL!="dm-[0-9]*", GOTO="end_mpath"''')
	rules.append('''SYSFS{device/model}}== "2810XIV" GOTO="end_mpath"''')
	rules.append('''PROGRAM!="/sbin/mpath_wait %M %m", GOTO="end_mpath"''')
	rules.append('''ACTION=="add", RUN+="/sbin/dmsetup ls --target multipath --exec '/sbin/kpartx -a -p p' -j %M -m %m"''')
	rules.append('''PROGRAM=="/sbin/dmsetup ls --target multipath --exec /bin/basename -j %M -m %m", RESULT=="?*", NAME="%k", SYMLINK="mpath/%c", OPTIONS="last_rule"''')
	rules.append('''PROGRAM!="/bin/bash -c '/sbin/dmsetup info -c --noheadings -j %M -m %m | /bin/grep -q .*:.*:.*:.*:.*:.*:.*:part[0-9]*-mpath-'", GOTO="end_mpath"''')
	rules.append('''PROGRAM=="/sbin/dmsetup ls --target linear --exec /bin/basename -j %M -m %m", NAME="%k", RESULT=="?*", SYMLINK="mpath/%c", OPTIONS="last_rule"''')
	rules.append('''LABEL="end_mpath"''')
	if not udev_install_rule(number=40, name='multipath', rulesList=rules):
		return 0
	return 1

def multipath_configure_udev_rules():
	if not udev_install_rule_45():
		return 0 
	if not udev_install_rule_40():
		return 0
	return 1

def is_rule_45_configured():
	output = execute_and_log('cat /etc/udev/rules.d/45-ibm-xiv.rules | head -n 1')
	if not is_error_free(output):
		return 0
	return 1

def is_rule_40_configured():
	output = execute_and_log('cat /etc/udev/rules.d/40-multipath.rules | head -n 1')
	if not is_error_free(output):
		return 0
	return 1

def is_multipathconf_configured():
	output = execute_and_log('cat /etc/multipath.conf | grep 2810XIV')
	if not is_error_free(output):
		return 0
	if output == '':
		return 0
	return 1

def is_getuid_configured():
	output = execute_and_log('cat ' + GETUID_CALLOUT_PATH + ' | head -n 1')
	if not is_error_free(output):
		return 0
	return 1

def is_multipath_configured():
	if not is_rule_45_configured():
		return 0
	if not is_rule_40_configured():
		return 0
	if not is_multipathconf_configured():
		return 0
	if not is_getuid_configured():
		return 0
	return 1

def is_multipath_kernel_module_loaded():
	output = execute_and_log('lsmod | grep multipath | head -n 1')
	if not is_error_free(output):
		return 0
	return 1

def load_kernel_modules():
	output = execute_and_log('insmod  /lib/modules/`uname -r`/kernel/drivers/md/dm-multipath.ko')
	if not is_error_free(output) and not output.__contains__('File exists') and not output.__contains__('Unknown symbol in module'):
		logger.error('failed to load kernel module dm-multipath')
		return 0
	output = execute_and_log('insmod  /lib/modules/`uname -r`/kernel/drivers/md/multipath.ko')
	if not is_error_free(output) and not output.__contains__('File exists') and not output.__contains__('Unknown symbol in module'):
		logger.error('failed to load kernel module multipath')
		return 0
	output = execute_and_log('insmod  /lib/modules/`uname -r`/kernel/drivers/md/dm-round-robin.ko')
	if not is_error_free(output) and not output.__contains__('File exists') and not output.__contains__('Unknown symbol in module'):
		logger.error('failed to load kernel module round-robin')
		return 0
	return 1

def multipath_scan():
	output = execute_and_log('multipath')
	if output.__contains__('error calling out /sbin/scsi_id'):
		return 1
	if not is_error_free(output):
		return 0
	return 1

def multipath_start():
	if not load_kernel_modules():
		return 0
	output = execute_and_log('/etc/init.d/multipathd start')
	if not is_error_free(output):
		return 0
	return multipath_scan()

def multipathd_service_healthcheck():
        output = execute_and_log('/etc/init.d/multipathd status')
        if not is_error_free(output):
                logger.error('multipathd daemon does not exists')
	        return 0
        if not output.__contains__('running'):
                logger.warning('multipathd daemon is not in running state')
	output = execute_and_log('chkconfig --list multipathd')
	if not is_error_free(output):
		logger.error('multipathd daemon is not configured properly')
		return 0
	if not output.__contains__('3:on'):
		logger.error('multipathd daemon is not configured properly')
		return 0		
	return 1

def multipath_healthcheck():
	if not is_multipath_configured():
		return 0
	if not is_multipathconf_configured():
		return 0
	return multipathd_service_healthcheck()
	if not is_multipath_kernel_module_loaded():
		logger.warn('multipath kernel module are not loaded, loading them now')
		return load_kernel_modules()
	return 1

def multipath_verify():
	return 	multipath_healthcheck()

def multipathd_service_config():
	output = execute_and_log('chkconfig --add multipathd')
	output = execute_and_log('chkconfig --level 345 multipathd on')
	return 1

def multipath_configure():
	if not is_multipath_configured():
		logger.info('configuring multipath')
	if not multipath_configure_udev_rules():
		return 0
	if not multipath_conf_install():
		return 0
	if not multipathd_service_config():
		return 0		
	logger.info('multipath is configured')
	return multipath_start()

